@@ -8,7 +8,7 @@
  *
  * Copyright (c) 2021-2023: MigOps Inc
  * Copyright (c) 2023: Gilles Darold
- * Copyright (c) 2024: HexaCluster Corp
+ * Copyright (c) 2024-2025: HexaCluster Corp
  *
  *-------------------------------------------------------------------------
  */
@@ -192,14 +192,16 @@ static void pghist_shmem_request(void);
 static void pghist_shmem_startup(void);
 static void pgph_shmem_startup(void);
 static void pgaf_shmem_startup(void);
+#if PG_VERSION_NUM >= 120000
 static int  entry_cmp(const void *lhs, const void *rhs);
+#endif
 static Size pgph_memsize(void);
 static void pg_password_history_internal(FunctionCallInfo fcinfo);
 static void fix_log(ErrorData *edata);
 static Size pgaf_memsize(void);
 static void credcheck_max_auth_failure(Port *port, int status);
 static float get_auth_failure(const char *username, Oid userid, int status);
-static float save_auth_failure(const char *username, Oid userid);
+static float save_auth_failure(Port *port, Oid userid);
 static void remove_auth_failure(const char *username, Oid userid);
 static void pg_banned_role_internal(FunctionCallInfo fcinfo);
 
@@ -215,6 +217,7 @@ static char *username_contain = NULL;
 static bool username_contain_password = true;
 static bool username_ignore_case = false;
 static char *username_whitelist = NULL;
+static char *max_auth_whitelist = NULL;
 
 /* Password flags*/
 static int password_min_length = 1;
@@ -245,7 +248,7 @@ char *str_to_sha256(const char *str, const char *salt);
 #endif
 
 bool check_whitelist(char **newval, void **extra, GucSource source);
-bool is_in_whitelist(char *username);
+bool is_in_whitelist(char *username, char *whitelist);
 
 static char *to_nlower(const char *str, size_t max) {
   char *lower_str;
@@ -506,27 +509,29 @@ check_whitelist(char **newval, void **extra, GucSource source)
 
 /* check if the username is in the whitelist */
 bool
-is_in_whitelist(char *username)
+is_in_whitelist(char *username, char *whitelist)
 {
 	char       *rawstring;
 	List       *elemlist;
 	ListCell   *l;
-	int len =  strlen(username_whitelist);
+	int len =  0;
 
 	Assert(username != NULL);
+	Assert(whitelist != NULL);
 
+	len =  strlen(whitelist);
 	if (len == 0)
 		return false;
 
 	/* Need a modifiable copy of string */
 	rawstring = palloc0(sizeof(char) * (len+1));
-	strcpy(rawstring, username_whitelist);
+	strcpy(rawstring, whitelist);
 	/* Parse string into list of identifiers */
 	if (!SplitIdentifierString(rawstring, ',', &elemlist))
 	{
 		ereport(ERROR,
 				(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
-				errmsg("username list is invalid: %s", username_whitelist)));
+				errmsg("username list is invalid: %s", whitelist)));
 		list_free(elemlist);
 		pfree(rawstring);
 		return false;
@@ -551,6 +556,7 @@ is_in_whitelist(char *username)
 	return false;
 }
 
+
 static void password_check(const char *username, const char *password)
 {
 
@@ -935,6 +941,7 @@ rename_user_in_history(const char *username, const char *newname)
 /*
  * qsort comparator for sorting into increasing usage order
  */
+#if PG_VERSION_NUM >= 120000
 static int
 entry_cmp(const void *lhs, const void *rhs)
 {
@@ -948,6 +955,7 @@ entry_cmp(const void *lhs, const void *rhs)
         else
                 return 0;
 }
+#endif
 
 static void
 remove_password_from_history(const char *username, const char *password, int numentries)
@@ -1243,7 +1251,7 @@ check_password(const char *username, const char *password,
 #ifdef USE_CRACKLIB
 			const char *reason;
 #endif
-			if (is_in_whitelist((char *)username))
+			if (is_in_whitelist((char *)username, username_whitelist))
 				break;
 
 			statement_has_password = true;
@@ -1329,6 +1337,11 @@ _PG_init(void)
 				NULL,
 				NULL);
 
+	DefineCustomStringVariable(
+				"credcheck.whitelist_auth_failure",
+				gettext_noop("comma separated list of username to exclude from max authentication failure check"), NULL,
+				&max_auth_whitelist, "", PGC_SUSET, 0, check_whitelist, NULL, NULL);
+
 #if PG_VERSION_NUM < 150000
 	EmitWarningsOnPlaceholders("credcheck");
 
@@ -1401,7 +1414,7 @@ cc_ProcessUtility(PEL_PROCESSUTILITY_PROTO)
 			/* We only take care of user renaming */
 			if (stmt->renameType == OBJECT_ROLE && stmt->newname != NULL)
 			{
-				if (is_in_whitelist(stmt->newname) || is_in_whitelist(stmt->subname))
+				if (is_in_whitelist(stmt->newname, username_whitelist) || is_in_whitelist(stmt->subname, username_whitelist))
 					break;
 			
 				/* check the validity of the username */
@@ -1421,10 +1434,10 @@ cc_ProcessUtility(PEL_PROCESSUTILITY_PROTO)
 			ListCell      *option;
 			char          *password;
 			bool           save_password = false;
-			DefElem    *dpassword = NULL;
 			DefElem    *dvalidUntil = NULL;
+			DefElem    *dpassword = NULL;
 
-			if (is_in_whitelist(stmt->role->rolename))
+			if (is_in_whitelist(stmt->role->rolename, username_whitelist))
 				break;
 
 			/* Extract options from the statement node tree */
@@ -1467,9 +1480,11 @@ cc_ProcessUtility(PEL_PROCESSUTILITY_PROTO)
 							errmsg(gettext_noop("the VALID UNTIL option must NOT have a date beyond %d days"), password_valid_max)));
 			}
 
+#if PG_VERSION_NUM >= 120000
 			/* The password can be saved into the history */
 			if (save_password)
 				save_password_in_history(stmt->role->rolename, password);
+#endif
 			break;
 		}
 
@@ -1485,7 +1500,7 @@ cc_ProcessUtility(PEL_PROCESSUTILITY_PROTO)
 			DefElem    *dpassword = NULL;
 			DefElem    *dvalidUntil = NULL;
 
-			if (is_in_whitelist(stmt->role))
+			if (is_in_whitelist(stmt->role, username_whitelist))
 				break;
 
 			/* check the validity of the username */
@@ -1543,8 +1558,10 @@ cc_ProcessUtility(PEL_PROCESSUTILITY_PROTO)
 					(errcode(ERRCODE_INVALID_AUTHORIZATION_SPECIFICATION),
 						errmsg(gettext_noop("require a VALID UNTIL option with a date beyond %d days"), password_valid_max)));
 			/* The password can be saved into the history */
+#if PG_VERSION_NUM >= 120000
 			if (save_password)
 				save_password_in_history(stmt->role, password);
+#endif
 			break;
 		}
 
@@ -2196,6 +2213,14 @@ credcheck_max_auth_failure(Port *port, int status)
 	if (status != STATUS_OK)
 		pg_usleep(1000L * auth_delay_milliseconds);
 
+	/* get out if the user is whitelisted for auth failure */
+	if (is_in_whitelist(port->user_name, max_auth_whitelist))
+	{
+		if (prev_ClientAuthentication)
+			prev_ClientAuthentication(port, status);
+		return;
+	}
+
 	/* check for max auth failure */
 	if (fail_max > 0 && status != STATUS_EOF)
 	{
@@ -2207,7 +2232,7 @@ credcheck_max_auth_failure(Port *port, int status)
 
 			/* register the auth failure if we not reach allowed max failure */
 			if (status == STATUS_ERROR && fail_num <= fail_max)
-				fail_num = save_auth_failure(port->user_name, userOid);
+				fail_num = save_auth_failure(port, userOid);
 
 			/* reject, this account has been banned */
 			if (fail_num >= fail_max)
@@ -2268,16 +2293,18 @@ get_auth_failure(const char *username, Oid userid, int status)
 }
 
 static float
-save_auth_failure(const char *username, Oid userid)
+save_auth_failure(Port *port, Oid userid)
 {
 	pgafHashKey key;
 	pgafEntry  *entry;
-	float fail_cnt = 0.5;
+	float fail_cnt = 1;
 
-	if (!EnableSSL)
-		fail_cnt = 1;
+	/*
+	if (port->ssl_in_use)
+		fail_cnt = 0.5;
+	*/
 
-	Assert(username != NULL);
+	Assert(port->user_name != NULL);
 
 	if (fail_max == 0)
 		return 0;
@@ -2296,15 +2323,17 @@ save_auth_failure(const char *username, Oid userid)
 	entry = (pgafEntry *) hash_search(pgaf_hash, &key, HASH_FIND, NULL);
 	if (entry)
 	{
-		if (EnableSSL)
+		/*
+		if (port->ssl_in_use)
 			fail_cnt = entry->failure_count + 0.5;
 		else
+		*/
 			fail_cnt = entry->failure_count + 1;
 
-		elog(DEBUG1, "Remove entry in auth failure hash table for user %s", username);
+		elog(DEBUG1, "Remove entry in auth failure hash table for user %s", port->user_name);
 		hash_search(pgaf_hash, &entry->key, HASH_REMOVE, NULL);
 	}
-	elog(DEBUG1, "Add new entry in auth failure hash table for user %s (%d, %f)", username, userid, fail_cnt);
+	elog(DEBUG1, "Add new entry in auth failure hash table for user %s (%d, %f)", port->user_name, userid, fail_cnt);
 
 	/* OK to create a new hashtable entry */
 	entry = pgaf_entry_alloc(&key, fail_cnt);
@@ -2463,7 +2492,7 @@ pg_banned_role_internal(FunctionCallInfo fcinfo)
 		memset(values, 0, sizeof(values));
 		memset(nulls, 0, sizeof(nulls));
 
-		values[i++] = Int8GetDatum(entry->key.roleid);
+		values[i++] = ObjectIdGetDatum(entry->key.roleid);
 		values[i++] = Int8GetDatum(entry->failure_count);
 		if (entry->banned_date)
 			values[i++] = TimestampTzGetDatum(entry->banned_date);
